import del from 'del'
import esbuild from 'esbuild'
import * as fs from 'fs'
import _ncp from 'ncp'
import path from 'path'
import {promisify} from 'util'
import ugly from 'uglify-js'

const config = {	// default configurations that can still be changed by users
	log: {error: true},
	basePath: 'web',
	distPath: 'dist',
	straightCopy: [],
	copyAssets: true,
}

const ncp = promisify(_ncp)

function verbose(...args) {
	config.log.info(...args)
}

function verror(...args) {
	config.log.error(...args)
}

function buildScript(basePath, distPath, file) {
	verbose(`Building script ${file}`)
	const clear = esbuild.buildSync({
		bundle: true,
		entryPoints: [path.join(basePath, file)],
		external: ['*webn', '*extern-html.js'],
		format: 'esm',
		outfile: path.join(distPath, file),
		treeShaking: true,
		write: false,
	})
	const {path: filePath, text} = clear.outputFiles[0]
	const uglyCode = ugly.minify(text.replace(/import\s*?['"].*?extern-html\.js['"]/g, ''))
	return uglyCode.code
}

function buildHTML(basePath, distPath, file, level = 0) {
    if(level > 100) throw 'Too many levels'
    return fs.readFileSync(path.join(basePath, file), 'utf8')
        .replace(/<extern-html((?:\s+?\w+=".+?")*)><\/extern-html>/g, (match, props) => {
			const attributes = [...props.matchAll(/(\w+?)\s*?\=\s*?"(.+?)"/g)].map(([, attribute, value]) => ({attribute, value}))
			const src = attributes.find(({attribute}) => attribute.toLowerCase() == 'src')
			const onload = attributes.find(({attribute}) => attribute.toLowerCase() == 'onload')
			if(!src) throw new Error(`No src attribute found in extern-html tag: "${match}"`)
            return `${buildHTML(basePath, distPath, src.value, level + 1)}${onload && `<script type="module">${onload.value}</script>` || ''}`
        })
        .replace(/(?:<\s*link\s+href\s*=\s*"(.+?)"\s*rel=\s*"stylesheet"\s*>|<\s*link\s+rel=\s*"stylesheet"\s*href\s*=\s*"(.+?)"\s*>)/g, (_, file) => {
			if(file.startsWith('http')) return `<link href="${file}" rel="stylesheet">`
			return `<style>${fs.readFileSync(path.join(basePath, file), 'utf8')}</style>`
        })
        .replace(/(?:<\s*script\s+src\s*=\s*"(.+?)"\s*type=\s*"module"\s*>\s*<\s*\/script\s*>|<\s*script\s+type=\s*"module"\s*src\s*=\s*"(.+?)"\s*>\s*<\s*\/script\s*>)/g, (tag, file) => {
			return `<script type="module">${buildScript(basePath, distPath, file)}</script>`
        })
  }

async function build() {
	try {
		verbose(`Building to ${config.distPath}`)
		await del(config.distPath)
		await fs.promises.mkdir(config.distPath)
		await Promise.all(config.straightCopy.map(async s => {
			verbose(`Copying asset ${s}`)
			const dist = path.join(config.distPath, config.copyAssets && s.startsWith('assets') ? s.replace('assets', '') : s)
			await ncp(path.join(config.basePath, s), dist)
		}))
		verbose('Creating index.html')
		fs.writeFileSync(path.join(config.distPath, 'index.html'), buildHTML(config.basePath, config.distPath, 'index.html'))
	}
	catch(error) {
		verror(error.message)
		verror(error.stack)
		// await del(config.distPath)
		throw error
	}
}

function configure(applicationConfig) {
	Object.assign(config, applicationConfig)
	if(config.log === true) config.log = {info: true, error: true}
	if(config.log === false) config.log = {info: false, error: false}
	if(typeof config.log == 'function') config.log = {info: config.log, error: config.log}
	if(config.log.info === true) config.log.info = (...args) => console.info(`[${new Date().toLocaleString()}]`, ...args)
	if(!config.log.info) config.log.info = () => {}
	if(applicationConfig?.log && !applicationConfig.log.hasOwnProperty('error')) config.log.error = true
	if(config.log.error === true) config.log.error = (...args) => console.error(`[${new Date().toLocaleString()}]`, ...args)
	if(config.log.error === false) config.log.error = () => {}
	if(config.copyAssets && fs.readdirSync(config.basePath).includes('assets')) {
		fs.readdirSync(path.join(config.basePath, 'assets')).map(asset => path.join('assets', asset)).forEach(asset => config.straightCopy.push(asset))
	}
}

export default async function(applicationConfig) {
	configure(applicationConfig)
	console.log(config.straightCopy)
	await build()
}
